@@ -1,6 +1,8 @@
 import os
-import crypt
 import hashlib
+import base64
+import getpass
+
 from odict import odict
 from plumber import (
     plumber,
@@ -35,9 +37,9 @@ from node.ext.ugm import (
 
 class FileStorage(Storage):
     """Storage behavior handling key/value pairs in a file.
-    
+
     Cannot contain node children. Useful for node attributes stored in a file.
-    
+
     XXX: extend node.behaviors.common.NodeChildValidate by ``allow_node_childs``
          attribute.
     """
@@ -45,13 +47,13 @@ class FileStorage(Storage):
     unicode_keys = default(True)
     unicode_values = default(True)
     delimiter = default(':')
-    
+
     @override
     def __init__(self, name=None, parent=None, file_path=None):
         self.__name__ = name
         self.__parent__ = parent
         self.file_path = file_path
-    
+
     @default
     @property
     def storage(self):
@@ -60,7 +62,7 @@ class FileStorage(Storage):
             if self.file_path and os.path.isfile(self.file_path):
                 self.read_file()
         return self._storage_data
-    
+
     @default
     def read_file(self):
         data = self._storage_data
@@ -72,7 +74,7 @@ class FileStorage(Storage):
                 if not isinstance(v, unicode) and self.unicode_values:
                     v = v.decode('utf-8')
                 data[k] = v.strip('\n')
-    
+
     @default
     def write_file(self):
         lines = list()
@@ -85,11 +87,13 @@ class FileStorage(Storage):
                 k = k.encode('utf-8')
             if isinstance(v, unicode) and self.unicode_values:
                 v = v.encode('utf-8')
+            if v is None:
+                v = ''
             line = self.delimiter.join([k, v]) + '\n'
             lines.append(line)
         with open(self.file_path, 'w') as file:
             file.writelines(lines)
-    
+
     @default
     def __call__(self):
         self.write_file()
@@ -106,7 +110,7 @@ class FileAttributes(object):
 
 
 class UserBehavior(BaseUserBehavior):
-    
+
     @default
     def user_data_attributes_factory(self, name=None, parent=None):
         user_data_dir = os.path.join(parent.data_directory, 'users')
@@ -114,35 +118,35 @@ class UserBehavior(BaseUserBehavior):
             os.mkdir(user_data_dir)
         user_data_path = os.path.join(user_data_dir, parent.name)
         return FileAttributes(name, parent, user_data_path)
-    
+
     attributes_factory = default(user_data_attributes_factory)
-    
+
     @override
     def __init__(self, name=None, parent=None, data_directory=None):
         self.__name__ = name
         self.__parent__ = parent
         self.data_directory = data_directory
-    
+
     @default
     @locktree
     def __call__(self, from_parent=False):
         self.attrs()
         if not from_parent:
             self.parent.parent.attrs()
-    
+
     @default
     def add_role(self, role):
         self.parent.parent.add_role(role, self)
-    
+
     @default
     def remove_role(self, role):
         self.parent.parent.remove_role(role, self)
-    
+
     @default
     @property
     def roles(self):
         return self.parent.parent.roles(self)
-    
+
     @default
     @property
     def groups(self):
@@ -152,7 +156,7 @@ class UserBehavior(BaseUserBehavior):
             if self.name in group.member_ids:
                 ret.append(group)
         return ret
-    
+
     @default
     @property
     def group_ids(self):
@@ -176,7 +180,7 @@ class User(object):
 
 
 class GroupBehavior(BaseGroupBehavior):
-    
+
     @default
     def group_data_attributes_factory(self, name=None, parent=None):
         group_data_dir = os.path.join(parent.data_directory, 'groups')
@@ -184,73 +188,73 @@ class GroupBehavior(BaseGroupBehavior):
             os.mkdir(group_data_dir)
         group_data_path = os.path.join(group_data_dir, parent.name)
         return FileAttributes(name, parent, group_data_path)
-    
+
     attributes_factory = default(group_data_attributes_factory)
-    
+
     @override
     def __init__(self, name=None, parent=None, data_directory=None):
         self.__name__ = name
         self.__parent__ = parent
         self.data_directory = data_directory
-    
+
     @default
     def __getitem__(self, key):
         return self.parent.parent.users[key]
-    
+
     @default
     @locktree
     def __delitem__(self, key):
         if not key in self.member_ids:
             raise KeyError(key)
         self._remove_member(key)
-    
+
     @default
     def __iter__(self):
         for id in self.member_ids:
             yield id
-    
+
     @default
     @locktree
     def __call__(self, from_parent=False):
         self.attrs()
         if not from_parent:
             self.parent.parent.attrs()
-    
+
     @default
     def add(self, id):
         if not id in self.member_ids:
             self._add_member(id)
-    
+
     @default
     def add_role(self, role):
         self.parent.parent.add_role(role, self)
-    
+
     @default
     def remove_role(self, role):
         self.parent.parent.remove_role(role, self)
-    
+
     @default
     @property
     def roles(self):
         return self.parent.parent.roles(self)
-    
+
     @default
     @property
     def users(self):
         return [self.parent.parent.users[id] for id in self.member_ids]
-    
+
     @default
     @property
     def member_ids(self):
         return [id for id in self.parent.storage[self.name].split(',') if id]
-    
+
     @default
     def _add_member(self, id):
         member_ids = self.member_ids
         member_ids.append(id)
         member_ids = sorted(member_ids)
         self.parent.storage[self.name] = ','.join(member_ids)
-    
+
     @default
     def _remove_member(self, id):
         member_ids = self.member_ids
@@ -271,7 +275,7 @@ class Group(object):
 
 
 class SearchBehavior(Behavior):
-    
+
     @default
     def _compare_value(self, term, value):
         # XXX: this should be done by regular expressions.
@@ -294,7 +298,7 @@ class SearchBehavior(Behavior):
         if term == value:
             return True
         return False
-    
+
     @default
     def search(self, criteria=None, attrlist=None,
                exact_match=False, or_search=False):
@@ -357,9 +361,11 @@ class SearchBehavior(Behavior):
 
 
 class UsersBehavior(SearchBehavior, BaseUsersBehavior):
-    
+
     unicode_values = default(False)
-    
+    salt_len = default(8)
+    hash_func = default(hashlib.sha256)
+
     @override
     def __init__(self, name=None, parent=None,
                  file_path=None, data_directory=None):
@@ -368,7 +374,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         self.file_path = file_path
         self.data_directory = data_directory
         self._mem_storage = dict()
-    
+
     @override
     def __getitem__(self, key):
         if not key in self.storage:
@@ -380,7 +386,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
                 name=key, parent=self, data_directory=self.data_directory)
             self._mem_storage[key] = user
             return user
-    
+
     @override
     @locktree
     def __setitem__(self, key, value):
@@ -388,7 +394,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         if not key in self.storage:
             self.storage[key] = ''
         self._mem_storage[key] = value
-    
+
     @override
     @locktree
     def __delitem__(self, key):
@@ -399,7 +405,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         del self._mem_storage[key]
         if key in self.parent.attrs:
             del self.parent.attrs[key]
-    
+
     @override
     @locktree
     def __call__(self, from_parent=False):
@@ -408,7 +414,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
             value(True)
         if not from_parent:
             self.parent.attrs()
-    
+
     @default
     def create(self, id, **kw):
         user = User(name=id, parent=self, data_directory=self.data_directory)
@@ -416,12 +422,12 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
             user.attrs[k] = v
         self[id] = user
         return user
-    
+
     @default
     def id_for_login(self, login):
         # XXX
         return login
-    
+
     @default
     def authenticate(self, id=None, pw=None):
         if not id in self.storage:
@@ -430,7 +436,7 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         if not self.storage[id]:
             return False
         return self._chk_pw(pw, self.storage[id])
-    
+
     @default
     def passwd(self, id, oldpw, newpw):
         if not id in self.storage:
@@ -439,18 +445,15 @@ class UsersBehavior(SearchBehavior, BaseUsersBehavior):
         if self.storage[id]:
             if not self._chk_pw(oldpw, self.storage[id]):
                 raise ValueError(u"Old password does not match.")
-        self.storage[id] = crypt.crypt(newpw, self._get_salt(id))
-    
-    @default
-    def _get_salt(self, id):
-        hash = hashlib.md5()
-        hash.update(id)
-        return hash.digest()[:2]
-    
+        salt = os.urandom(self.salt_len)
+        hashed = base64.b64encode(self.hash_func(newpw + salt).digest() + salt)
+        self.storage[id] = hashed
+
     @default
     def _chk_pw(self, plain, hashed):
-        salt = hashed[:2]
-        return hashed == crypt.crypt(plain, salt)
+        hashed = base64.b64decode(hashed)
+        salt = hashed[-self.salt_len:]
+        return hashed == self.hash_func(plain + salt).digest() + salt
 
 
 class Users(object):
@@ -467,7 +470,7 @@ class Users(object):
 
 
 class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
-    
+
     @override
     def __init__(self, name=None, parent=None,
                  file_path=None, data_directory=None):
@@ -476,7 +479,7 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
         self.file_path = file_path
         self.data_directory = data_directory
         self._mem_storage = dict()
-    
+
     @override
     def __getitem__(self, key):
         if not key in self.storage:
@@ -488,7 +491,7 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
                 name=key, parent=self, data_directory=self.data_directory)
             self._mem_storage[key] = group
             return group
-    
+
     @override
     @locktree
     def __setitem__(self, key, value):
@@ -496,7 +499,7 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
         if not key in self.storage:
             self.storage[key] = ''
         self._mem_storage[key] = value
-    
+
     @override
     @locktree
     def __delitem__(self, key):
@@ -506,7 +509,7 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
         id = 'group:%s' % key
         if id in self.parent.attrs:
             del self.parent.attrs[id]
-    
+
     @override
     @locktree
     def __call__(self, from_parent=False):
@@ -515,7 +518,7 @@ class GroupsBehavior(SearchBehavior, BaseGroupsBehavior):
             value(True)
         if not from_parent:
             self.parent.attrs()
-    
+
     @default
     def create(self, id, **kw):
         group = Group(name=id, parent=self, data_directory=self.data_directory)
@@ -539,15 +542,15 @@ class Groups(object):
 
 
 class UgmBehavior(BaseUgmBehavior):
-    
+
     @default
     def role_attributes_factory(self, name=None, parent=None):
         attrs = FileAttributes(name, parent, parent.roles_file)
         attrs.delimiter = '::'
         return attrs
-    
+
     attributes_factory = default(role_attributes_factory)
-    
+
     @override
     def __init__(self,
                  name=None,
@@ -562,7 +565,7 @@ class UgmBehavior(BaseUgmBehavior):
         self.groups_file = groups_file
         self.roles_file = roles_file
         self.data_directory = data_directory
-    
+
     @override
     def __getitem__(self, key):
         if not key in self.storage:
@@ -573,49 +576,49 @@ class UgmBehavior(BaseUgmBehavior):
                 self['groups'] = Groups(file_path=self.groups_file,
                                       data_directory=self.data_directory)
         return self.storage[key]
-    
+
     @override
     @locktree
     def __setitem__(self, key, value):
         self._chk_key(key)
         self.storage[key] = value
-    
+
     @override
     def __delitem__(self, key):
         raise NotImplementedError(u"Operation forbidden on this node.")
-    
+
     @override
     def __iter__(self):
         for key in ['users', 'groups']:
             yield key
-    
+
     @override
     @locktree
     def __call__(self):
         self.attrs()
         self.users(True)
         self.groups(True)
-    
+
     @default
     @property
     def users(self):
         return self['users']
-    
+
     @default
     @property
     def groups(self):
         return self['groups']
-    
+
     @default
     def roles(self, principal):
         id = self._principal_id(principal)
         return self._roles(id)
-    
+
     @default
     @property
     def roles_storage(self):
         return self.attrs
-    
+
     @default
     @locktree
     def add_role(self, role, principal):
@@ -625,7 +628,7 @@ class UgmBehavior(BaseUgmBehavior):
         roles.append(role)
         roles = sorted(roles)
         self.attrs[self._principal_id(principal)] = ','.join(roles)
-    
+
     @default
     @locktree
     def remove_role(self, role, principal):
@@ -635,21 +638,21 @@ class UgmBehavior(BaseUgmBehavior):
         roles.remove(role)
         roles = sorted(roles)
         self.attrs[self._principal_id(principal)] = ','.join(roles)
-    
+
     @default
     def _principal_id(self, principal):
         id = principal.name
         if isinstance(principal, Group):
             id = 'group:%s' % id
         return id
-    
+
     @default
     def _roles(self, id):
         attrs = self.attrs
         if not id in attrs:
             return list()
         return [role for role in attrs[id].split(',') if role]
-    
+
     @default
     def _chk_key(self, key):
         if not key in ['users', 'groups']:
@@ -667,4 +670,4 @@ class Ugm(object):
         DefaultInit,
         Nodify,
         OdictStorage,
-    )
\ No newline at end of file
+    )
